//******************************************************************************
// jacobi-seq.c
//
// Jacobi-Iteration fuer die Laplace-Gleichung Laplace u = f,
// sequentielle Variante mit finiten Differenzen (FD). Konsistenz-Ordnung ist
// 2, Jacobi konvergiert aber extrem langsam fuer n >= 64 (eventuell euklid.
// statt Max.-Norm verwenden)
//
// 06.12.2011
//******************************************************************************
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpi.h>

const double Pi = 3.1415926535897932385;

// Ecke oben rechts
const double eor = 1.0;

// Quelle f
double f(double x1, double x2)
{
  return 2.0 * Pi * Pi * sin(Pi*x1) * sin(Pi*x2);
}

// Auswerten der analytischen Loesung
double aSol(double x1, double x2, double sol, double err)
{
  return fabs(sol - sin(Pi*x1) * sin(Pi*x2));
}

// Auswerten des Fehlers in der Maximums-Norms an allen Gitterpunkten
double maxNorm(double* sol, long n)
{
  long i, j;
  double maxErr = 0.0;
  const double h  = eor / (n - 1);

  for (i=0; i<n; i++)
  {
    for (j=0; j<n; j++)
    {
      const double x1 = i * h;
      const double x2 = j * h;
      double tmp = fabs(sol[i*n+j] - sin(Pi*x1) * sin(Pi*x2));
      if (maxErr < tmp)
        maxErr = tmp;
    }
  }

  return maxErr;
}

// Maximums-Norm des Residuums
double maxNormRes(const double* const x, const double* const b, double** A, long m)
{
  double max = 0.0;
  long i,j;

  for (i=0; i<m; i++)
  {
    double res = b[i];
    for (j=0; j<m; j++)
      res -= A[i][j] * x[j];
    if (max < fabs(res))
      max  = fabs(res);
  }

  return max;
}

// Initialisierung der Matrix und Vektoren
void init(double* x, double* b, double** A, long m, long n)
{
  const double h  = eor / (n - 1);
  const double h2 = h * h;
  long i, j, k;

  // Iteriere über alle Gitter-Punkte
  for (i=0; i<n; i++)
  {
    for (j=0; j<n; j++)
    {
      // Zeilen-Index
      const long zi = i * n + j;

      // Initialisiere Loesungs-Vektor x mit 0
      x[zi] = 0.0;

      // Initialisiere Last-Vektor b mit f(x,y)
      const double x1 = i * h;
      const double x2 = j * h;
      b[zi] = f(x1, x2);

      // Generiere Matrix-Zeile fuer Gitterpunkt (i,j)
      for (k=0; k<m; k++)
      {
        // Setze Zeile zurueck
        A[zi][k] = 0.0;

        // Diagonal-Eintrag
        A[zi][zi] = 4.0 / h2;

        // Neben-Diagonal-Eintraege
        if ((k == zi-1) || (k == zi+1) || (k == zi-n) || (k == zi+n))
          A[zi][k] = -1.0 / h2;

        // Setze Randbedingungen
        if ((i == 0) || (i == n-1) || (j == 0) || (j == n-1))
        {
          x[zi] = 0.0;
          b[zi] = 0.0;

          A[zi][k]  = 0.0;
          A[zi][zi] = 1.0;
        }
      }
    }
  }
}

// Output einer Matrix
void outputMatrix(double** A, long m)
{
  long i, j;
  for (i=0; i<m; i++)
  {
    for (j=0; j<m; j++)
    {
      fprintf(stdout, "%.4e  ", A[i][j]);
    }
    fprintf(stdout, "\n");
  }
  fprintf(stdout, "\n\n");
}

// Output eines Vektors
void output(double* v, long n)
{
  long i, j;
  const double h  = eor / (n - 1);
  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
    {
      const double x1 = i * h;
      const double x2 = j * h;
      fprintf(stdout, "%.4e  %.4e  %.4e\n", x1, x2, v[i*n+j]);
    }
  fprintf(stdout, "\n\n");
}

// Jacobi-Schritt
void jacobi(double* xneu, const double* const xalt, const double* const b, double** A, long m)
{
  long i=0, j=0;

  for (i=0; i<m; ++i)
  {
    xneu[i] = 0.0;

    // Berechne Skalar-Produkt
    for (j=0; j<m; ++j)
    {
      xneu[i] += A[i][j] * xalt[j];
    }

    // Subtrahiere den Fall i = j
    xneu[i] -= A[i][i] * xalt[i];

    // Finish des Jacobi-Schritts
    xneu[i] *= -1.0;
    xneu[i] += b[i];
    if (A[i][i] == 0)
    {
      fprintf(stderr, "Division by 0\n.");
      exit (1);
    }
    else
      xneu[i] /= A[i][i];
  }
}

//*****************************************************************************
// main
//*****************************************************************************
int main(int argc, char** argv)
{
  const long maxIt = 10000; // max. Iterationen des Verfahrens
  long it;                  // Schleifenzaehler
  long n;                   // Problemgroesse in einer Richtung

  double*  x;               // Unbekannte
  double*  y;               // alte Unbekannte
  double*  b;               // rechte Seite
  double** A;               // Matrix

  const double eps = 1e-5;  // Abbruchbedingung
  double diff;              // Maximumsnorm des Abstandes zweier Loesungen
  double t = 0.0;           // fuer die Zeitmessung

  // Lese Laenge des Unbekannten-Vektors x
  if (argc != 2)
  {
    fprintf(stderr, "Syntax: %s <n>\n", argv[0]);
    exit (1);
  }
  n = atoi(argv[1]);

  // allokiere Speicher fuer A, x, y und b
  const long m = n * n;
  x = (double *)  malloc(sizeof(double)  * m); // alter Vektor x
  y = (double *)  malloc(sizeof(double)  * m); // neuer Vektor y
  b = (double *)  malloc(sizeof(double)  * m); // rechte Seite b
  A = (double **) malloc(sizeof(double*) * m); // Zeilen Matrix A
  long i;
  for (i=0; i<m; i++)
    A[i] = (double*) malloc(sizeof(double) * m);

  if (A == NULL || b == NULL || y == NULL || x == NULL)
  {
    fprintf(stderr, "Could not allocate memory\n");
    return 1;
  }

  // Initialisiere A, b und x
  init(x, b, A, m, n);

  // Jacobi-Schritte
  t = MPI_Wtime();
  for (it=1; it<((long) (maxIt/2)+1); it++)
  {
    // Trick um das Umkopieren x^{m} = y (neuer Vektor in alten Vektor)
    // Vektorzu sparen: 2 Jacobi-Schritte, Loesung x^(m) ist dann abwechselnd
    // in x oder y. d.h. wir machen insgesamt 2*it Jacobi-Schritte)
    jacobi(y, x, b, A, m);
    jacobi(x, y, b, A, m);

    // Konvergenz-Check und Ende Zeitmessung
    diff = maxNormRes(x, b, A, m);
    fprintf(stdout, "Norm(res): %.16e\n", diff);
    if (diff < eps)
    {
      t = MPI_Wtime() - t;
      //fprintf(stdout, "Iter: %ld, t: %f s, Norm(res): %.16e\n", (2*it), t, diff);
      break;
    }
  }

  // Output finaler Loesungsvektor und Fehler in Max-Norm
  output(x, n);
  printf("Max. Error: %e\n", maxNorm(x, n));

  // Speicher-Freigabe
  free(x);
  free(y);
  free(b);
  for (i=0; i<m; i++)
    free(A[i]);
  free(A);

  return (0);
}
